@@ -7,6 +7,7 @@ import os
 import re
 import time
 from collections import namedtuple
+from collections import deque
 from datetime import datetime
 
 import attrs
@@ -195,10 +196,12 @@ class CourseMetadata:
 # noinspection PyTypeChecker,PyUnresolvedReferences
 class CanvasRobot(object):
     db: callable
+    from collections import deque
     TOT_WEIGHT: int = 100
     year: int = AC_YEAR
-    def __init__(self, reset_api_keys=False, years_back = 0):
+    def __init__(self, reset_api_keys=False, years_back = 0, msg_queue=None):
         self.year = AC_YEAR - years_back
+        self.queue = msg_queue
         config = CanvasConfig(reset_api_keys=reset_api_keys)
         self.canvas = canvasapi.Canvas(config.url, config.api_key)
         self.admin = self.canvas.get_account(config.admin_id) if config.admin_id else None
@@ -213,6 +216,8 @@ class CanvasRobot(object):
         self.errors = []
         self.actions = []
 
+    def add_to_queue(self, msg, value):
+        self.queue.put((msg,value))
     def lookup_teachers_db(self):
         db = self.db
         teachers = db((db.course2user.role == 'T') &
@@ -306,7 +311,7 @@ class CanvasRobot(object):
             # print("There are {} modules in page".format(len(page)))
         return
 
-    def course_metadata(self, course_id, examination_names=None):
+    def course_metadata(self, course_id, examination_names=None, show_candidates=False):
         """
         return dict with metadata of this course
         as a dict.
@@ -332,33 +337,50 @@ class CanvasRobot(object):
         pages = filter(lambda p: p.title[0:3] != 'UVT', pages)
         nr_pages = len(list(pages))
         assignments = course.get_assignments()
-        assignments_summary = "Assignments (online_uploads have details)\n"
+        assignments_summary = "Assignments:\n"
         examination_candidates = []
         for assignment in assignments:
+            if assignment.name in examination_names:
+                assignments_summary += f"Examination assignment: {assignment.name}"
+            else:
+                assignments_summary += f"Examination candidate: {assignment.name}" \
+                    if show_candidates else ""
+
+                candidate = ExaminationDTO(course_id,
+                                           course.name,
+                                           assignment.name)
+
+                examination_candidates.append(candidate)
             submissions_summary = ""
             submissions = assignment.get_submissions()
-            for submission in submissions:
-                if submission.submission_type == "online_upload":
+            try:
+                for idx, submission in enumerate(submissions, start=1):
                     if assignment.name in examination_names:
-                        submissions_summary += "Examination:"
-                    else:
-                        submissions_summary+="Possible Examination:"
+                        if submission.submission_type == "online_upload":
+                            submissions_summary += (f"({idx}. {submission.submission_type}) graded "
+                                                    f"{submission.grade} at "
+                                                    f"{submission.graded_at}. "
+                                                    f"Checked for plagiarism: "
+                                                    f"{submission.has_originality_report if hasattr(submission,'has_originality_report') else 'no has_originality_report attribute!'}\n"
+                                                    )
+                        else:
+                            submissions_summary+=(f"({idx}. {submission.submission_type}) graded "
+                                                  f"{submission.grade} at "
+                                                  f"{submission.graded_at}\n")
+            except BaseException as exc:
+                logging.exception(f"In Course:{course_id} {course.name} "
+                                  f"Assignment {assignment.name} "
+                                  f"Submission nr {idx} type{submission.type} {exc}")
+                raise
 
-                        candidate = ExaminationDTO(course_id,
-                                                   course.name,
-                                                   assignment.name)
 
-                        examination_candidates.append(candidate)
-                    submissions_summary+=(f"/ngraded "
-                                          f"{submission.grade} at "
-                                          f"{submission.graded_at}")
-            assignments_summary+=assignment.name + ":"
-            assignments_summary+=submissions_summary
+            assignments_summary+=f"\n{submissions_summary}"
         nr_assignments = len(list(assignments))
         quizzes = course.get_quizzes()
         nr_quizzes = len(list(quizzes))
         files = course.get_files()
         nr_files = len(list(files))
+        # check for uploaded examination files
         examinations_summary = ""
         examination_files = 0
         for file in files:
@@ -367,9 +389,9 @@ class CanvasRobot(object):
                 examination_files+=1
                 examinations_summary+=f"\n{file.display_name}"
         if examination_files == 0:
-            examinations_summary += f"\nNo examination_files in {EXAMINATION_FOLDER}"
+            examinations_summary += f"\nNo examination files in folder {EXAMINATION_FOLDER}"
         else:
-            examinations_summary += f"\nTotal: {examination_files}"
+            examinations_summary += f"\nTotal: {examination_files} examination files"
         # was this working earlier 2.2.0 ?
         # try:
         #    collaborations = course.get_collaborations()
@@ -935,13 +957,27 @@ class CanvasRobot(object):
 
     def get_examinations_from_database(self, candidate=False):
         db = self.db
-        records = db((db.examination.id>0)&
-                     (db.examination.candidate==candidate)
-                     ).select(db.examination.ALL)
+        # include the NULL values
+        qry = ((db.examination.id>0)&
+               (db.examination.candidate==candidate)) if candidate else (db.examination.id > 0)
+        records = db( qry ).select(db.examination.ALL)
         return records
-    def get_courses_from_database(self):
+    def get_courses_from_database(self,
+                                  skip_courses_without_students=False,
+                                  qry=None,
+                                  orderby=None):
         db = self.db
-        records = db(db.course.id>0).select(db.course.ALL)
+        if skip_courses_without_students:
+            cur_qry = (db.course.nr_students>0)
+        else:
+            cur_qry = (db.course.id>0)
+
+        if qry:
+            cur_qry = cur_qry & qry
+        orderby=orderby or db.course.course_code
+        records = db(cur_qry).select(db.course.ALL,
+                                     orderby=orderby)
+
         return records
     def get_course_from_database(self,course_id):
         db = self.db
@@ -956,13 +992,13 @@ class CanvasRobot(object):
     def update_record_db(self, search_field, search_id, table ,field, value):
         db = self.db
         ud_fields = {field: value}
-        row = db(db[table][search_field] == search_id).select()
-        row2 = db(db.course.course_id == search_id).select()
+        #row = db(db[table][search_field] == search_id).select()
+        #row2 = db(db.course.course_id == search_id).select()
         result = db(db[table][search_field] == search_id).update(**ud_fields)
         db.commit()
 
     # noinspection PyUnusedLocal
-    def update_database_from_canvas(self, single_course=None):
+    def update_database_from_canvas(self, single_course=None, max_number=None, stop_list=None):
         """
             Using the canvasapi to read the list of TST courses and
             - record internal_id course_id, fname and instructors in the
@@ -976,19 +1012,29 @@ class CanvasRobot(object):
 
         db = self.db
         examinations = self.get_examinations_from_database(candidate=False)
-
-
-        logging.info(f'open courselist for year {self.year} - {self.year+1}')
+        msg= f'open courselist for year {self.year} - {self.year+1}'
+        self.add_to_queue(msg,None)
+        logging.info(msg)
         num_rows = 0
         # tst = self.canvas.get_account(6)  # admin account
         courses=[self.get_course(single_course),] if single_course else self.admin.get_courses()
+        num_courses = len(list(courses))
+        max_number = max_number or num_courses
+
+        for idx,course in enumerate(courses):
+            self.add_to_queue(course.name, (idx,num_courses))
 
-        for course in courses:
             # only insert/update course if current year unless single_course
             if (str(course.sis_course_id)[:4] != str(self.year)
                     or course.name.endswith('conclude')) and not single_course:
                 continue
-            logging.info("course: {}".format(course.name))  # course
+            if course.name in (stop_list or []):
+                continue
+            if idx > max_number:
+                break
+            logging.debug("course: {}".format(course.name))  # course
+            students = course.get_users(enrollment_type="student")
+            nr_students=len(list(students))
             teachers = course.get_users(enrollment_type="teacher")
             teachers_ids = []
             for teacher in teachers:
@@ -1024,7 +1070,8 @@ class CanvasRobot(object):
             format_str = "%Y-%m-%dT%H:%M:%SZ"
             creation_date = datetime.strptime(course.created_at, format_str)
             course_id = None
-            canonical_examination_names = [row.name for row in examinations if row.course==course_id]
+            canonical_examination_names = [row.name for row in examinations if (row.course==course.id
+                                                                                and not row.candidate)]
             md = self.course_metadata(course.id, canonical_examination_names)
             try:
                 course_id = db.course.update_or_insert(db.course.course_id == course.id,
@@ -1036,6 +1083,7 @@ class CanvasRobot(object):
                                                        name=course.name,
                                                        creation_date=creation_date,
                                                        ac_year=self.year,
+                                                       nr_students=nr_students,
                                                        nr_modules=md.nr_modules,
                                                        nr_module_items=md.nr_module_items,
                                                        nr_pages=md.nr_pages,
@@ -1053,11 +1101,13 @@ class CanvasRobot(object):
             else:
                 course_id = db(db.course.course_id == course.id).select().first().id
             for cand in md.examination_candidates:
+                # candidate is True if course_name in examination_list else False
                 db.examination.update_or_insert((db.examination.course == cand.course_id)&
                                                  (db.examination.name == cand.name),
                                                  course=cand.course_id,
                                                  course_name=cand.course_name,
-                                                 name=cand.name)
+                                                 name=cand.name,
+                                                )
             for user_id in teachers_ids:
                 db.course2user.update_or_insert((db.course2user.course == course_id) &
                                                 (db.course2user.user == user_id),
